/* **************************************************************************

  flexray.h, a library supporting trees-of-lookup-tables.

Copyright (C) 2000 John Conner Atkeson

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

For further information, contact:
John Conner Atkeson III
jcatkeson at gmail dot com
http://jca3.freeshell.org
3959 Persimmon Drive #103 Fairfax Va. 22031

************************************************************************** */

#ifndef flexray_h
#define flexray_h 1

#include <math.h>
#include "rosetta.h"

#ifndef tin_bool
  #define tin_bool unsigned char
  #define true 0x1
  #define false 0x0
#endif
#ifndef int8
  #define int8 char
#endif
#ifndef uint8
  #define uint8 unsigned int8
#endif
#ifndef int32
  #define int32 int
#endif
#ifndef uint32
  #define uint32 unsigned int32
  #define maxuint32 (~(uint32)0)
#endif

#define testing_performance 0

#if testing_performance
typedef struct leafT
/* *************************************************************
************************************************************* */
{
  int value;
  double xyz[3];
}leafT;
#else
typedef void leafT;
#endif

typedef struct flexelT
/* *************************************************************
*  Flex array element.
************************************************************* */
{
  uint8 num_children;
  union down{
    struct flexelT *next;/* the bar (array) I point to. */
    leafT *leaf;/* leaf node. */
  }down;
}flexelT;

/* generic routine for disposing of user-defined leaves. */
typedef void flexray_leafkillT(void *leaf);

flexelT *flexray_create();
void flexray_init2(flexelT *topnode);
tin_bool flexray_insert(flexelT *topnode,uint32 key,leafT *newleaf);
void flexray_destroy(flexelT *topnode,flexray_leafkillT *leafkill);
leafT *flexray_lookup(flexelT *topnode,uint32 key);
leafT *flexray_remove(flexelT *topnode,uint32 key);
leafT *flexray_toggle(flexelT *topnode,uint32 key,leafT *newleaf);

#endif // #ifndef flexray_h
