/* **********************************************************
*  Platform portability library (after the Rosetta stone).
*  All ugliness goes here.
********************************************************** */

#ifdef __cplusplus
  extern "C" {
#endif
#ifndef rosetta_h 
  #define rosetta_h 1

  #include <stdio.h>
  #include <stdlib.h>
  #include <math.h>
  #include <string.h>

  #include <fcntl.h>
  #include <sys/types.h>

  /* if glib.h is unavailable, you'll have to set this to 0 by hand. */
  #define HAVE_GLIB_H 1

  #if HAVE_GLIB_H
    // #include <glib.h>
  #endif

  #ifdef _MSC_VER  /* check for MSVC */
    #include <io.h>
    #include <direct.h>
    #include <SYS\STAT.H>
    // #include <new.h> 

    #define strcasecmp stricmp
    #define strncasecmp _strnicmp
    #define S_ISDIR(a) (_S_IFDIR & a) /* only directories */
    #define S_ISREG(a) (_S_IFREG & a) /* only files */
    /* todo: also support _S_IEXEC */
  
    #define dir_handleT long /* whole dir handle */
    #define dir_entryT struct _finddata_t /* single entry type */
  #else
    /* Let gtk-win32 handle dirent under Windows. */
    #ifndef G_OS_WIN32
      #include <dirent.h>
    #endif

    /* whole dir handle */
    #define dir_handleT DIR *
    /* single entry type */
    #define dir_entryT struct dirent *

    #include <sys/stat.h>
  #endif

  #ifdef __TURBOC__
    #define S_ISREG(x) (S_IFREG & x)
    #define strcasecmp strcmpi
    #include <ERRNO.H>
  #else
    #ifndef _MSC_VER  /* check for MSVC */
      #include <unistd.h>
    #endif
  #endif

  #ifdef _WIN32
    #define slash "\\"
    #define slashch '\\'
  #else
    #define slash "/"
    #define slashch '/'
  #endif
  #ifndef __DJGPP
    #ifndef O_BINARY
      #define O_BINARY 0
    #endif
    #define uclock clock
    #define uclock_t clock_t
    #define UCLOCKS_PER_SEC 1
  #endif
  #ifndef S_IRUSR
    #define S_IRUSR  00400 /* read by owner */
    #define S_IWUSR  00200 /* write by owner */
    #define S_IXUSR  00100 /* execute/search by owner */
  #endif
  #ifndef S_IWOTH  /* or S_IXGRP? */
    #define S_ISUID  04000 /* set user ID on execution */
    #define S_ISGID  02000 /* set group ID on execution */
    #define S_ISVTX  01000 /* sticky bit */
    #define S_IRGRP  00040 /* read by group */
    #define S_IWGRP  00020 /* write by group */
    #define S_IXGRP  00010 /* execute/search by group */
    #define S_IROTH  00004 /* read by others */
    #define S_IWOTH  00002 /* write by others */
    #define S_IXOTH  00001 /* execute/search by others */
  #endif

  #define datpermissions (S_IWUSR | S_IRUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH)
  #define dirpermissions (datpermissions | S_IXUSR | S_IXGRP | S_IXOTH)

  /* some numeric types and essential constants */

  #ifndef int8
    #define int8 char
    #define uint8 unsigned int8
    #define maxuint8 (~(uint8)0)/* biggest possible 8-bit integer. */
    #define maxint8 (maxuint8>>1)
  #endif
  #ifndef int16
    #define int16 short
    #define uint16 unsigned int16
    #define maxuint16 (~(uint16)0)/* biggest possible 16-bit integer. */
    #define maxint16 (maxuint16>>1)
  #endif
  #ifndef int32
    #define int32 long
    #define uint32 unsigned int32
    #define maxuint32 (~(uint32)0)/* biggest possible 32-bit integer. */
    #define maxint32 (maxuint32>>1)
  #endif
  #ifndef int64
    #define int64 long long
    #define uint64 unsigned int64
    #define maxuint64 (~(uint64)0)
    #define maxint64 (maxuint64>>1)
  #endif
  /* This integer is intended to scale with system bit size. */
  #ifndef sysint
    #define sysint int
    #define sysuint unsigned sysint
    #define maxsysuint (~(sysuint)0)
    #define maxsysint (maxsysuint>>1)
  #endif
  #if 0
  #if (sizeof(sysuint)<sizeof(uint64))
    #define biguint uint64
  #else
    #define biguint sysuint
  #endif
  #endif

  /* **********************************************************
  *  Integer size macros.
  ********************************************************** */
  /* size in bits */
  #define bitsize(val) (sizeof(val)*8)
  /* maximum unsigned value for this type */
  #define maxuval(val) (~( (~(sysuint)0) <<bitsize(val) ))
  /* maximum signed value for this type */
  #define maxval(val) (maxuval(val)>>1)
  /* as far as you can shift a 1 without going off the top end. */
  #define shiftrange(val) (bitsize(val)-1)
  /* convert signed to unsigned, with same range, by inverting sign bit. */
  #define s2u(val) (((unsigned)(val)) ^ (1<<shiftrange(val)))
  /* convert unsigned to signed, also same range. */
  #define u2s(val) ((signed)(  (val)^(1<<shiftrange(val))  ))

  /* Whatever is the largest array-indexing integer that we can use on this system. */
  #ifndef intdex 
    #define intdex sysint
  #endif
  #ifndef refint
    #define refint sysint
  #endif
  #ifndef busint
    #define busint sysint
  #endif
  #ifndef byte
    #define byte uint8
  #endif
  #ifndef maxlong
    #define maxlong maxint32
  #endif
  #ifndef ulong
    #define ulong unsigned long
  #endif
  #ifndef __cplusplus
    #ifndef bool
      #define bool uint8
      #define true (1==1)
      #define false (1==0)
    #endif
  #endif

    #ifndef bool
      #define bool uint8
      #define true (1==1)
      #define false (1==0)
    #endif

  #ifndef TRUE
    #define TRUE true
  #endif
  #ifndef FALSE
    #define FALSE false
  #endif

  #ifndef PI
    #define PI 3.14159265358979
  #endif

  #define twopi (PI*2.0)
  #define fudge (1.0e-200)/* to workaround divide-by-zero etc. */
  #define infinity 1.0/fudge;

  /* **********************************************************
  *  C/C++ compatible memory allocation.
  ********************************************************** */
  #if false
    #define allocsafe(num_units,unit_type) ((unit_type*)calloc(num_units,sizeof(unit_type)))
    #define reallocsafe(original,num_units,unit_type) ((unit_type*)realloc(original,(num_units)*sizeof(unit_type)))
    #define freesafe(ptr) if (ptr!=NULL){free(ptr); ptr=NULL;}
  #else
    // #define NULL (void*)
    // #define NULL (0)
    #define allocsafe(num_units,unit_type) (new unit_type[num_units])
    #define reallocsafe(original,num_units,unit_type) ((unit_type*)realloc(original,(num_units)*sizeof(unit_type)))
    #define freesafe(ptr) {}// if (ptr!=NULL){free(ptr); ptr=NULL;}
  #endif


  /* **********************************************************
  *   strlistT is for directory listing support.
  *  Especially for case-insensitive file searching.
  ********************************************************** */
  typedef struct {
    sysint num;
    char **list;
  }strlistT;

#if false
  /* ***********************************************************
  *********************************************************** */

  char *dos2unix(char *path);
  char *unix2dos(char *path);
  /*  Convert path slashes for appropriate operating system. */
  void syspath(char *path);
  char LowCh(char inch);
  void Upc(char *s);
  void Lwc(char *instr);
  char *strrstr(char *s1,char *s2);
  bool EndMatch(const char *s1,char *s2);
  void rosetta_trimstr(char *instring);

  /* **********************************************************  
  *  End of misc. string utilities.
  ********************************************************** */ 

  /* **********************************************************
  * start directory listing support.
  ********************************************************** */
  bool open_dir(const char *dirname,dir_handleT *dir_handle,dir_entryT *dir_entry);
  bool read_dir(dir_handleT dir_handle,dir_entryT *entry);
  void close_dir(dir_handleT dir_handle);
  char *dir_entry_name(dir_entryT *entry);

  bool loaddirlist(const char *dirpath,strlistT *list);
  void freedirlist(strlistT *list);
  size_t searchdirlist(strlistT *list,const char *key);
  bool get_fname_nocase(const char *searchdir,char *target,char *results);

#endif

  #if false
    junkyard
  /*
    printf("errno:%i\n",errno);
    printf("ENOENT:%i\n",ENOENT);
    printf("ENOMEM:%i\n",ENOMEM);
    printf("EBADF:%i\n",EBADF);
  */
  #endif
#endif /* ifndef rosetta_h */

#ifdef __cplusplus
}
#endif
