// This is the main DLL file.
// tinlib.cpp : Defines the initialization routines for the DLL.
/* **************************************************************************

  TIN.C, a library to create Triangulated Irregular Networks
from xyz point data.

Copyright (C) 2000 John Conner Atkeson III

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

For further information, contact:
John Conner Atkeson III
jcatkeson at gmail dot com
http://jca3.freeshell.org
3959 Persimmon Drive #103 Fairfax Va. 22031

************************************************************************** */

#include "stdafx.h"
#include "tinlib.h"

time_t randseed;

#define testing 0

char *vertexpnt="vertexpnt";

namespace tinlib {

tin_bool point_test(XYZType *Pt)
/* *******************************************************************
*  Just a stub to fill the user-definable filter.
* ******************************************************************* */
{
  return(true);
}

lineT *line_create(uint32 pdex1,uint32 pdex2,lstateT state)
/* *******************************************************************
* ******************************************************************* */
{
  lineT *line;
  line=allocsafe(1,lineT);
  line->pdex1=pdex1; line->pdex2=pdex2; line->state=state;
  return(line);
}/* line_create */
void line_free(lineT *line)
/* *******************************************************************
*  Free the memory allocated to a line.
* ******************************************************************* */
{
  freesafe(line);
}/* line_free */

void DrawLoop(brickptrT loop_wall,plist_class *points,int color)
/* *******************************************************************
*  For debugging, not used.
* ******************************************************************* */
{
  XYZType *pt1,*pt2,*P;
  intdex pdex1,pdex2,brickcnt;
  brickptrT now;
  lineT *this_line;

  P=points->P;

  brickcnt=0;
  now=loop_wall;
  do{
    this_line=now->linptr;
    pdex1=this_line->pdex1; pdex2=this_line->pdex2;
    pt1=&(P[pdex1]); pt2=&(P[pdex2]);

    if (this_line->state==outer){
      color=1;
    }else{
      color=2;
    }
#if allegro
    line(screen,
      pt1->Loc[0],pt1->Loc[1], pt2->Loc[0],pt2->Loc[1], color);
#endif
    now=now->next;
    brickcnt++;
//    color++;
  }while(now!=loop_wall);
}/* DrawLoop */

void draw_triangle(plist_class *points,intdex pdex1,intdex pdex2,intdex pdex3,int color)
/* *******************************************************************
* ******************************************************************* */
{
  XYZType *pt1,*pt2,*pt3;

//  color=1;
//  color%=60; if (color==0)color=1;

  pt1=&(points->P[pdex1]); pt2=&(points->P[pdex2]); pt3=&(points->P[pdex3]);
#if allegro
  line(screen,
    pt1->Loc[0],pt1->Loc[1], pt2->Loc[0],pt2->Loc[1], color);

  line(screen,
    pt3->Loc[0],pt3->Loc[1], pt2->Loc[0],pt2->Loc[1], color);

  line(screen,
    pt1->Loc[0],pt1->Loc[1], pt3->Loc[0],pt3->Loc[1], color);
#endif
}/* draw_triangle */

void tinner_class::draw_triangles()
/* *******************************************************************
*  Draw triangles onscreen, only used for debugging with allegro.
* ******************************************************************* */
{
  triT *tri;
  uint32 tcnt;//,num_triangles;

  // num_triangles=num_triangles;
  for(tcnt=0;tcnt<num_triangles;tcnt++){
    tri=(triT *)flexray_lookup(tri_tree,tcnt);
    draw_triangle(points,tri->vtex[0],tri->vtex[1],tri->vtex[2],4);
  }
}/* tin_draw_triangles */

uint32 count_bricks(brickptrT loop_wall)
/* *******************************************************************
* ******************************************************************* */
{
  intdex brickcnt;
  brickptrT now;

  brickcnt=0; now=loop_wall;
  do{
    now=now->next; brickcnt++;
  }while(now!=loop_wall);
  return(brickcnt);
}/* count_bricks */


void save_triangle(FILE *outfile,plist_class *points,intdex pdex0,intdex pdex1,intdex pdex2)
/* *******************************************************************
*  Save one triangle to a file. (Companion to tin_save_triangles.)
* ******************************************************************* */
{
  XYZType *pt0,*pt1,*pt2;
  pt0=&(points->P[pdex0]); pt1=&(points->P[pdex1]); pt2=&(points->P[pdex2]);

  fprintf(outfile,"(%1.1lf, %1.1lf, %1.1lf), ",pt0->Loc[0],pt0->Loc[1],pt0->Loc[2]);
  fprintf(outfile,"(%1.1lf, %1.1lf, %1.1lf), ",pt1->Loc[0],pt1->Loc[1],pt1->Loc[2]);
  fprintf(outfile,"(%1.1lf, %1.1lf, %1.1lf)\n",pt2->Loc[0],pt2->Loc[1],pt2->Loc[2]);
}/* save_triangle */

void tinner_class::save_triangles(char *filename)
/* *******************************************************************
*  Save all triangles to a file.
* ******************************************************************* */
{
  triT *tri;
  uint32 tcnt;//,num_triangles;
  FILE *outfile;

  outfile=fopen(filename,"w");

  // num_triangles=num_triangles;
  for(tcnt=0;tcnt<num_triangles;tcnt++){
    tri=(triT *)flexray_lookup(tri_tree,tcnt);
    save_triangle(outfile,points,tri->vtex[0],tri->vtex[1],tri->vtex[2]);
  }
  fclose(outfile);
}/* tin_save_triangles */

void tinner_class::save_tinfile(char *filename)
/* *******************************************************************
*  Save whole structure of TIN to a file, including connectivity.
* ******************************************************************* */
{
  plist_class *pl; XYZType *Plist,*P;
  uint32 pcnt,num_points;
  triT *tri;
  uint32 tcnt;//,num_triangles;
  FILE *outfile;
  char outstr[256];

  // num_triangles=num_triangles;

  pl=points; num_points=pl->NumPoints; Plist=pl->P;

  outfile=fopen(filename,"w");
  fprintf(outfile,"num_points=%u num_triangles=%u\n",num_points,num_triangles);
  /* Save raw data points, in order. */
  for(pcnt=0;pcnt<num_points;pcnt++){
    P=&(Plist[pcnt]);
    sprintf(outstr,"p %1.2lf %1.2lf %1.2lf",P->Loc[0],P->Loc[1],P->Loc[2]);
    fprintf(outfile,"%s\n",outstr);
  }
  /* Save traingles as indexes of data points. */
  for(tcnt=0;tcnt<num_triangles;tcnt++){
    tri=(triT *)flexray_lookup(tri_tree,tcnt);
    sprintf(outstr,"t %li %li %li",tri->vtex[0],tri->vtex[1],tri->vtex[2]);
    fprintf(outfile,"%s\n",outstr);
  }
  fclose(outfile);
}/* save_tinfile */

void tinner_class::store_triangle(intdex pdex0,intdex pdex1,intdex pdex2)
/* *******************************************************************
*  Store a triangle to memory in a flexray.
* ******************************************************************* */
{
  triT *tri;
  tin_bool success;

  tri=allocsafe(1,triT);
  tri->vtex[0]=pdex0; tri->vtex[1]=pdex1; tri->vtex[2]=pdex2;
  success=flexray_insert(tri_tree,num_triangles,tri);
  if(!success){
    printf("Error, a triangle has already been stored to this index!\n");
  }
  num_triangles++;
}/* store_triangle */

tin_bool tinner_class::get_triangle_full(full_triT *tri,uint32 tri_num)
/* *******************************************************************
*  Get a triangle with all its analogue coordinates.
* ******************************************************************* */
{
  triT *tridex;
  XYZType *P,*pt;
  int dimcnt,vcnt;

  tridex=(triT *)flexray_lookup(tri_tree,tri_num);

  if (tridex!=NULL){
    P=points->P;
    for (vcnt=0;vcnt<3;vcnt++){
      pt=&(P[tridex->vtex[vcnt]]);
      for (dimcnt=0;dimcnt<3;dimcnt++){
        tri->vtex[vcnt][dimcnt]=pt->Loc[dimcnt];
      }
    }
    return(true);/* True if triangle found */
  }else{
    return(false);/* False if triangle not found */
  }
}/* tinner_class::get_triangle_full */

void tinner_class::save_outline(char *filename)
/* *******************************************************************
*   Save the loop wall to a file.
* ******************************************************************* */
{
  // brickptrT loop_wall;
  // plist_class *points;
  XYZType *pt1,*pt2,*P;
  intdex pdex1,pdex2,brickcnt;
  brickptrT now;
  lineT *this_line;
  FILE *outfile;
  
  // loop_wall=loop_wall;
  // points=points;
  
  outfile=fopen(filename,"w");

  P=points->P;

  brickcnt=0;
  now=loop_wall;
  do{
    this_line=now->linptr;
    pdex1=this_line->pdex1; pdex2=this_line->pdex2;
    pt1=&(P[pdex1]); pt2=&(P[pdex2]);

/*    fprintf(outfile,"%1.2lf, %1.2lf\n ",pt1->Loc[0],pt1->Loc[1]); */

    fprintf(outfile,"%1.2lf %1.2lf %1.2lf - ",pt1->Loc[0],pt1->Loc[1],pt1->Loc[2]);
    fprintf(outfile,"%1.2lf %1.2lf %1.2lf\n",pt2->Loc[0],pt2->Loc[1],pt2->Loc[2]);

    now=now->next;
    brickcnt++;
  }while(now!=loop_wall);

  fclose(outfile);
}/* tin_save_outline */


/* *******************************************************************
*  Begin line list support.
* ******************************************************************* */

linelistT *linelist_create(int nlines)
/* *******************************************************************
* ******************************************************************* */
{
  int cnt;
  linelistT *lines;

  lines=allocsafe(1,linelistT); 

  lines->nlines=nlines;
  lines->L=allocsafe(nlines,lineptrT);

  /* probably never use this part. */
  for(cnt=0;cnt<nlines;cnt++){
    lines->L[cnt]=line_create(0,0,alive);
  }
  return(lines);
}
void linelist_destroy(linelistT *lines)
/* *******************************************************************
* ******************************************************************* */
{
  int cnt,nlines=lines->nlines;
  for(cnt=0;cnt<nlines;cnt++){
    freesafe(lines->L[cnt]); lines->L[cnt]=NULL;
  }
  freesafe(lines->L); freesafe(lines);
}/* linelist_destroy */
void leaf_destroy(void *lines)
/* *******************************************************************
* ******************************************************************* */
{
  linelist_destroy((linelistT *)lines);
}/* leaf_destroy */

void linelist_addone(linelistT *lines,lineT *line)
/* *******************************************************************
* ******************************************************************* */
{
  int last=lines->nlines;
  lines->nlines++;
  lines->L=reallocsafe(lines->L,lines->nlines,lineT*);
  lines->L[last]=line;
}
lineT *linelist_remove(linelistT *lines,int linenum)
/* *******************************************************************
* ******************************************************************* */
{
  int cnt,nlines;
  lineT *retline;

//  freesafe(lines->L[linenum]);
  retline=lines->L[linenum];
  lines->nlines--; nlines=lines->nlines;
  for (cnt=linenum;cnt<nlines;cnt++){
    lines->L[cnt]=lines->L[cnt+1];
  }
  lines->L=reallocsafe(lines->L,nlines,lineT*);
  return(retline);
}/* linelist_remove */
void linelist_pack(linelistT *lines,int nlines)
/* *******************************************************************
* ******************************************************************* */
{
  lines->L=reallocsafe(lines->L,nlines,lineT*);
  lines->nlines=nlines;
}
/* *******************************************************************
*  End line list support.
* ******************************************************************* */

void linetree_lookup(flexelT *line_tree,lineT *line,char *tag)
/* *******************************************************************
*  Remove one line from the linetree.
* ******************************************************************* */
{
  tin_bool found=false;
  linelistT *lines;
  lineT **L;
  int linenum,nlines;

  lines=(linelistT *)flexray_lookup(line_tree,line->pdex1);
  if (lines!=NULL){
    nlines=lines->nlines;
    L=lines->L;
  
    linenum=0;
    while(linenum<nlines){
      if(L[linenum]==line){found=true; break;}
      linenum++;
    }
  
    if(found){
      printf("I see line in %s.\n",tag);
    }else{
      printf("Dont see line in %s.\n",tag);
    }
  }else{
    printf("Dont see linelist in %s.\n",tag);
  }
}/* linetree_lookup */

void linetree_insert(flexelT *line_tree,lineT *line)
/* *******************************************************************
*  Add one line to the linetree.
* ******************************************************************* */
{
  tin_bool found=false,success;
  linelistT *lines;
  lines=(linelistT *)flexray_lookup(line_tree,line->pdex1);
  if(lines==NULL){
    lines=linelist_create(0);
    success=flexray_insert(line_tree,line->pdex1,lines);
    if(!success){
      printf("Error in linetree_insert, linelist already there.\n");
    }
  }
  linelist_addone(lines,line);
}/* linetree_insert */

void linetree_remove(flexelT *line_tree,lineT *line)
/* *******************************************************************
*  Remove one line from the linetree.
* ******************************************************************* */
{
  tin_bool found=false;
  linelistT *lines;
  lineT **L;
  int linenum,nlines;
  lineT *oldline;
  intdex pdex1;

  pdex1=line->pdex1;

  lines=(linelistT *)flexray_lookup(line_tree,pdex1);
  if (lines!=NULL){
    nlines=lines->nlines;
    L=lines->L;
  
    linenum=0;
    while(linenum<nlines){
      if(L[linenum]==line){found=true; break;}
      linenum++;
    }
  
    if(found){
      oldline=linelist_remove(lines,linenum);
      if(lines->nlines<=0){
        /* all lines are gone, remove the list. */
        lines=(linelistT *)flexray_remove(line_tree,pdex1);
        linelist_destroy(lines);
      }
      freesafe(oldline);
    }else{
      printf("Error in linetree_remove, line not found.\n");
    }
  }else{
    printf("Error in linetree_remove, linelist not found.\n");
  }
}/* linetree_remove */

void tri_destroy(void *triangle)
/* *******************************************************************
*  Free the memory allocated to a triangle.
* ******************************************************************* */
{
  freesafe((triT *)triangle);
//  freesafe(triangle);
}/* tri_destroy */


uint32 tinner_class::create_random(int numpoints)
/* *******************************************************************
*  Create a field of random points for testing.
* ******************************************************************* */
{
  int cnt,dimcnt;
  XYZType *pt;
  double randnum;

//  size=50;
//  randseed=955189213;
//  randseed=955234891;
  randseed=time(0);

  srand(randseed);
  for(cnt=0;cnt<8;cnt++){
    randnum=rand();
  }

  allocate_points(numpoints);

  for(cnt=0;cnt<numpoints;cnt++){
    for(dimcnt=0;dimcnt<3;dimcnt++){
      randnum=rand()%160+20;
      randnum+=((double)(rand()%12345))/12345.6;
//      randnum=sqrt(randnum)*17.0-40.0;

      points->P[cnt].Loc[dimcnt]=randnum;
    }
    pt=&(points->P[cnt]);
#if allegro
    line(screen,pt->Loc[0],pt->Loc[1],pt->Loc[0],pt->Loc[1], 0xff);
#endif
  }
  printf("rs:%li.",randseed);
  return(randseed);
}/* tin_create_random */

tin_bool in_salami(
  double x0,double y0,double x1,double y1,double radius,
  double xtarg,double ytarg)
/* **************************************************************
*  Returns true if xtarg,ytarg is found inside a sausage-shaped
* area.  Experimental, not currently used.
************************************************************** */
{
  tin_bool retval;
  double dx0,dy0,dx1,dy1,dxl,dyl;
  double rad00;
  double rad01;
  double linelen,dist,extension;

  dx0=(x0-xtarg); dy0=(y0-ytarg);
  rad00=sqrt(dx0*dx0 + dy0*dy0);
  if (rad00<=radius){/* Close enough to one endpoint. */
    retval=true;
//    printf("in radius0. ");
  }else{
    dx1=(x1-xtarg); dy1=(y1-ytarg);
    rad01=sqrt(dx1*dx1 + dy1*dy1);
    if (rad01<=radius){/* Close enough to the other endpoint. */
      retval=true;
//      printf("in radius1. ");
    }else{/* If all else failed, check for distance from centerline. */
      dxl=(x1-x0); dyl=(y1-y0);
      linelen=sqrt(dxl*dxl + dyl*dyl);
//      if (linelen==0.0) linelen=fudge;/* avoid divide-by-zero. */
//      dist=( ( (y0-ytarg)*dxl - (x0-xtarg)*dyl )/(linelen*linelen) )*linelen;
      dist=( (y0-ytarg)*dxl - (x0-xtarg)*dyl )/linelen;

      if (fabs(dist)<=radius){
        extension=((y0-ytarg)*(y0-y1)-(x0-xtarg)*(x1-x0))/(linelen*linelen);
        if (0.0<=extension && extension<=1.0){
          retval=true;
//          printf("near bar. ");
        }else{
          retval=false;
//          printf("outside end of bar. \n");
        }
      }else{
        retval=false;
      }
    }
  }
  return(retval);
}/* in_salami */

tin_bool tinner_class::FindClosest(intdex CenterDex,intdex *NbrDex,LogicType *Filter)
/* ***************************************************************
*  Pythagorean closest to point at CenterDex.
*************************************************************** */
{
  intdex zcnt,SubjIndex,min_point_index;
  double search_radius,MinDist,Dist;
  double centerX,centerY;
  // VectT *paramv;/* returns whatever traversezone found. */
  CubeT szone;
  tin_bool Found;

  centerX=points->P[CenterDex].Loc[0];
  centerY=points->P[CenterDex].Loc[1];
  min_point_index=-1; MinDist=maxint32;/* Index of nearest neighbor. */
  search_radius=1.0;/* search radius. */
  
  /* z stuff. */
  szone.L[0][2]=points->MinMax.L[0][2];  szone.L[1][2]=points->MinMax.L[1][2];
  
  Found=false;
  do{
    paramv->NumPoints=0;
    /* set search bounds. */
    SetCube(centerX-search_radius,centerY-search_radius,0.0,
            centerX+search_radius,centerY+search_radius,0.0,&szone);

    TraverseZone(sort,points,&szone,paramv);
    for (zcnt=0;zcnt<paramv->NumPoints;zcnt++){
      SubjIndex=paramv->V[zcnt];

      {/* start test distance part. */
        /* See if this point is closer than any other found. */
        XYZType *P=&(points->P[SubjIndex]);/* with-do for sample point. */
        Dist=Hypot(P->Loc[0]-centerX,P->Loc[1]-centerY);/* Pythagorean Distance */
        if(Dist<=search_radius){/* if within search circle. */
          if(SubjIndex!=CenterDex){/* if not myself. */
            if(Dist<MinDist){/* if closer than any other. */
              if( (*Filter)(P) ){/* user-defined filter         */
                min_point_index=SubjIndex; MinDist=Dist;
                Found=true;
              }
            }
          }
        }/* if inside radius */
      }/* end of test distance part. */

    }
    search_radius*=2.0;/* expand search zone for next. */
  }while (! (Found||(search_radius>=border_segment_max)) );
  (*NbrDex)=min_point_index;
  return(Found);
}/* End of FindClosest. */

tin_bool tinner_class::FindThird(intdex pt0,intdex pt1,intdex *MinSubjIndex,LogicType *Filter)
/* ***************************************************************
* Always searches on the left hand side of line travel
* direction.
*************************************************************** */
{
  intdex zcnt,datindex,min_point_index;
  double Tx0,Ty0,Tx1,Ty1;
  double HalfDeltaX,HalfDeltaY,MidX,MidY;
  double HalfBase,Angle,RadOffset,MaxAngle;
  double search_radius,Spread,Dist;
  double centerX,centerY;
  CubeT szone;
  XYZType *P,BaseP0,BaseP1;
  tin_bool Found;
  // VectT *paramv;/* returns whatever traversezone found. */

  min_point_index=-1;/* Index of nearest neighbor. */
  Tx0=points->P[pt0].Loc[0]; Ty0=points->P[pt0].Loc[1];
  Tx1=points->P[pt1].Loc[0]; Ty1=points->P[pt1].Loc[1];
  
  HalfDeltaX=(Tx1-Tx0)/2.0;HalfDeltaY=(Ty1-Ty0)/2.0;
  MidX=Tx0+HalfDeltaX;MidY=Ty0+HalfDeltaY;
  HalfBase=Hypot(HalfDeltaX,HalfDeltaY);/* starter radius */
  
  min_point_index=-1;/* to scare out bugs. */
  /* set actual base points. */
  BaseP0=points->P[pt0]; BaseP1=points->P[pt1];
  MaxAngle=-infinity;//maxint32;/* ridiculous starter comparison */

  Found=false;

  Angle=Pi/2.0;/* 90 degrees for starters, no obtuse angles */
  do{/* Search circle is defined by decreasing angle. */
    paramv->NumPoints=0;

    /* Radius of search circle */
    search_radius=HalfBase/sin(Angle); RadOffset=search_radius*cos(Angle);
    /* Center of search circle */
    centerX=MidX-(RadOffset *(HalfDeltaY/HalfBase));
    centerY=MidY+(RadOffset *(HalfDeltaX/HalfBase));
    SetCube(centerX-search_radius,centerY-search_radius,0.0,
            centerX+search_radius,centerY+search_radius,0.0,&szone);

    TraverseZone(sort,points,&szone,paramv);
    for (zcnt=0;zcnt<paramv->NumPoints;zcnt++){
      datindex=paramv->V[zcnt];
      {/* Triangular distance testing. */
        P=&(points->P[datindex]);/* with-do for sample point. */

        /* Pythagorean Dist. */
        Dist=Hypot(P->Loc[0]-centerX,P->Loc[1]-centerY);
        if(Dist<=search_radius){/* must be inside search circle to compete */

          /* looking for maximum obtuseness in angle. */
          Spread=AngleSpread(
            BaseP0.Loc[0],BaseP0.Loc[1],/* p0x, p0y */
            P->Loc[0],P->Loc[1],/* testx, testy */
            BaseP1.Loc[0],BaseP1.Loc[1]);/* p1x, p1y */

          if(Spread<=Pi){/* if greater than 180 degrees, ignore. */
            if((pt0!=datindex)&&(pt1!=datindex)){/* don't find selves. */
              if(Spread>MaxAngle){
                if( (*Filter)(P) ){/* user-defined filter */
                  if (1){//in_salami(Tx0,Ty0,Tx1,Ty1,border_segment_max,P->Loc[0],P->Loc[1])){
                    min_point_index=datindex;
                    MaxAngle=Spread;
                    Found=true;
                  }
                }
              }
            }
          }
#if 0
          { double arm1, arm2;
            arm1=Hypot(BaseP0.Loc[0]-P->Loc[0],BaseP0.Loc[1]-P->Loc[1]);
            arm2=Hypot(BaseP1.Loc[0]-P->Loc[0],BaseP1.Loc[1]-P->Loc[1]);
            if (arm1>arm2)maxarm=arm1; else maxarm=arm2;
          }
#endif
        }
      }/* end of triangular distance testing. */

    }
    Angle/=2.0;/* Angle gets narrower with each expansion of search.*/
  }while (! (Found || (RadOffset>=border_segment_max) || (Angle<Pi/500.0)) );
  (*MinSubjIndex)=min_point_index;
  return(Found);
}/* End of FindThird. */

void tinner_class::PushLine(lineT *newline)
/* *******************************************************************
*  Add a line to the line wall.
* ******************************************************************* */
{
  brickptrT Up,Novel,here;

  /* Add line to linetree. */
  linetree_insert(line_tree,newline);

  /* new brick */
  Novel = allocsafe(1,brickT);
  Novel->linptr=newline; /* create and set line pointer. */

  here=loop_wall;

  /* Insert the new link behind me. */
  Up=here->prev; /* save link to my predecessor */
  Up->next=Novel; /* set prev's down pointer to novel */
  here->prev=Novel; /* set my up pointer to novel */

  Novel->prev=Up;/* link novel prev to up */
  Novel->next=here;/* link novel prev to up */

  loop_wall=here;
}/* PushLine */

void tinner_class::PopLine()
/* *******************************************************************
* Take a line out of the line wall, the line pointed to by loop_wall.
* ******************************************************************* */
{
  brickptrT up,down,here;
  lineT *line;

  here=loop_wall;
  line=here->linptr; up=here->prev; down=here->next;
  up->next=down; down->prev=up;

  freesafe(here); here=down;
  linetree_remove(line_tree,line);
  loop_wall=here;
}/* PopLine */

#if 0
// deprecated.
void DisposeAll(brickptrT *Place)
/* *******************************************************************
*  Return the whole line loop to heap memory, along with the lines
* it points to.
* ******************************************************************* */
{
  brickptrT now;
  now=(*Place);
  while(now->next!=now){
    PopLine(&now);
  }
  linetree_remove(&line_tree,now->linptr); freesafe(now);
  (*Place)=NULL;
}/* DisposeAll */
#endif

brickptrT dispose_loop(brickptrT loop_wall)
/* *******************************************************************
*  Return the whole line loop to heap memory.
*  Disposes of the loop wall without touching the line tree.
* ******************************************************************* */
{
  brickptrT old,now;//,loop_wall;
  // loop_wall = *loop_wall_ptr;
  now=loop_wall;
  if (now!=NULL){
    do{
      old=now; now=now->next; freesafe(old);
    }while(now!=loop_wall);
  }
  // *loop_wall_ptr = NULL;
  return NULL;
}/* dispose_loop */

void kill_congruents(flexelT *line_tree,lineT *me)
/* *******************************************************************
* ******************************************************************* */
{
  linelistT *other_lines;
  lineT **L;
  uint32 pdex2,pdex1,nlines;
  uint32 lcnt;

  pdex2=me->pdex2; pdex1=me->pdex1;
  other_lines=(linelistT *)flexray_lookup(line_tree,pdex2);
  /* at this point, other_lines has returned a list of all
    the other lines that end where me-line begins. */
  if (other_lines!=NULL){
    nlines=other_lines->nlines; L=other_lines->L;
    for(lcnt=0;lcnt<nlines;lcnt++){
      /* if the pdex2 point is the same as my pdex1, we are congruent.
        So tag us both for extinction. */
      if (L[lcnt]->pdex2==pdex1){
        L[lcnt]->state=dead; me->state=dead;
      }
    }
  }
}/* kill_congruents */

tin_bool tinner_class::MakeFirst()
/* *******************************************************************
*  Make first seed triangle.
* ******************************************************************* */
{
  intdex num_points;
  intdex vert_index1=0, vert_index2=0, vert_index3=0;
  plist_class *pl;
  XYZType *Point1,*Point2,*P,*VertPt;
  tin_bool found=false;

  pl=points; num_points=pl->NumPoints;
  P=pl->P;

  vert_index1=0; vert_index2=0; vert_index3=0;
  /* this is just to pick a random point with neighbors. */
  while (vert_index1<num_points){    
    found=false;

    Point1 = &(P[vert_index1]);
    
    if (Point1->ref!=vertexpnt){
      /*
        check if point1 used here.  If used, skip looking for the
        other two points and look for the next point1.
  
        If we go all the way from beginning to end and find nothing,
        declare the tinning to be over.
      */
      /* Find nearest one that's not self */
      found=FindClosest(vert_index1,&vert_index2,&point_test);/* pythagorean closest. */
      if (found){
        Point2 = &(P[vert_index2]);
        /* Triangle search: Find point nearest to line. Always searches
        on the left hand side of line travel direction. */
        found=FindThird(vert_index1,vert_index2,&vert_index3,&point_test);
      }
    }
    /* repeat until one triangle can be completed, or we run out of points. */
    if (found)break;

    vert_index1++;
  }/* end while */
  if (!found) return(false);/* No more tinning can be done. */

  VertPt = &(P[vert_index3]);

  /* triangle is counter-clockwise */
  store_triangle(vert_index1,vert_index2,vert_index3);
  P[vert_index1].ref=vertexpnt;
  P[vert_index2].ref=vertexpnt;
  P[vert_index3].ref=vertexpnt;

  {/* with lines do make first triangle from lines. */
    lineT *newline;
    // brickptrT loop_wall_new;

    /* Add first line to linetree. */
    newline=line_create(vert_index3,vert_index2,alive);
    if (loop_wall==NULL){
      linetree_insert(line_tree,newline);
      loop_wall=allocsafe(1,brickT); 
	  loop_wall->linptr = newline;
      loop_wall->prev=loop_wall; loop_wall->next=loop_wall;/* points to self, a one-node ring. */
    }else{
      PushLine(newline);
    }
    /* Add second line to linetree. */
    newline=line_create(vert_index2,vert_index1,alive);
    PushLine(newline);

    /* Add third line - complete first triangle */
    newline=line_create(vert_index1,vert_index3,alive);
    PushLine(newline);
  }/* end with. */
  return(true);
}/* MakeFirst */

void tinner_class::allocate_points(int new_numpoints)
/* *******************************************************************
* ******************************************************************* */
{
  points = new plist_class(new_numpoints);
}/* allocate_points */
uint32 tinner_class::get_num_points()
/* *******************************************************************
* ******************************************************************* */
{
	if (points!=NULL){
	  return points->NumPoints;
	}else{
	  return 0;
	}
}/* tinner_class::get_num_points */

void tinner_class::set_point(ulong pdex,double xval,double yval,double zval)
/* *******************************************************************
* ******************************************************************* */
{
	XYZType *xyz = &(points->P[pdex]);
	xyz->Loc[0] = xval; xyz->Loc[1] = yval; xyz->Loc[2] = zval;
}/* set_point */

void tinner_class::get_point(ulong pdex, double% xval, double% yval, double% zval)
/* *******************************************************************
* ******************************************************************* */
{
	XYZType *xyz = &(points->P[pdex]);
	xval = xyz->Loc[0]; yval = xyz->Loc[1]; zval = xyz->Loc[2];
}/* get_point */

tin_bool tinner_class::get_triangle_indexes(ulong tdex, ulong% v0, ulong% v1, ulong% v2)
/* *******************************************************************
*  Return a triangle, as just the 3 indexes to the 3 corners in the 
* point array.
* ******************************************************************* */
{
	triT *tridex;
	tridex=(triT *)flexray_lookup(tri_tree,tdex);
	if (tridex!=NULL){
		v0 = tridex->vtex[0];
		v1 = tridex->vtex[1];
		v2 = tridex->vtex[2];
		return(true);/* True if triangle found */
	}else{
		return(false);/* False if triangle not found */
	}
}/* get_triangle_indexes */


tin_bool tinner_class::get_triangle_verts(int tdex,
									   double *x0,double *y0,double *z0, 
									   double *x1,double *y1,double *z1, 
									   double *x2,double *y2,double *z2)
/* *******************************************************************
* ******************************************************************* */
{
	tin_bool success;
	full_triT triangle; /* contains: double vtex[3][3], 3 xyz points */
	int vcnt;
	success = get_triangle_full(&triangle,tdex);
	if(success){

		vcnt = 0;
		*x0 = triangle.vtex[vcnt][0];
		*y0 = triangle.vtex[vcnt][1];
		*z0 = triangle.vtex[vcnt][2];

		vcnt = 1;
		*x1 = triangle.vtex[vcnt][0];
		*y1 = triangle.vtex[vcnt][1];
		*z1 = triangle.vtex[vcnt][2];

		vcnt = 2;
		*x2 = triangle.vtex[vcnt][0];
		*y2 = triangle.vtex[vcnt][1];
		*z2 = triangle.vtex[vcnt][2];

	}else{/* if it returned false, then we went out of range. */
	}
	return success;
}/* get_triangle_verts */


#if false
void tinner_class::tin_init()
/* *******************************************************************
*  Initialize, allocate, and sort TIN needs.
* ******************************************************************* */
{
  VectT *newsort = allocsafe(1,VectT);/* create vector headers. */
  VectT *newparamv = allocsafe(1,VectT);

  InitSortVect(newsort,newpoints->NumPoints);
  MakeBlankSortVect(newparamv,newpoints->NumPoints);

  jasort(newpoints,newsort);
  loop_wall=NULL;
  points=newpoints; sort=newsort; paramv=newparamv;
  num_triangles=0;
  /* default triangle leg size is practically infinity. */
  border_segment_max=1.0/fudge;
}/* tin_init */
#endif

void tinner_class::generate_tin(double new_border_segment_max)
/* *******************************************************************
* ******************************************************************* */
{
  intdex pdex1,pdex2,VertDex;
  tin_bool Done=false,found=false;
  lineT *ThisLine,*newline,*outermark;
  XYZType *Point1,*Point2;

  line_tree = flexray_create();
  tri_tree = flexray_create();

  if (true){// tin_init();
	  sort = sort_vect_create(points->NumPoints,true);/* create vector headers and fill them. */
	  paramv = sort_vect_create(points->NumPoints,false);

	  jasort(points,sort);
	  loop_wall=NULL;
	  num_triangles=0;
	  /* default triangle leg size is practically infinity. */
	  border_segment_max=1.0/fudge;
  }/* tin_init */

  /* set triangle leg size per the user/programmer`s request. */
  border_segment_max=new_border_segment_max;

  while (MakeFirst()){
    Done=false;
    do{
      ThisLine=loop_wall->linptr;
      kill_congruents(line_tree,ThisLine);
  
      if(ThisLine->state==dead){
        /* if a wall line is dead, don't grow on it. */
        outermark=NULL;
        if(loop_wall->next==loop_wall){
          Done=true;/* if wall is down to 1 brick, we're done. */
        }else{
          PopLine();/* Remove dead line from loop */
        }
      }else if (ThisLine->state==outer){
  
        /* Start checking if our *entire loop wall* is facing outer space. */
        if (outermark==NULL){
          outermark=ThisLine;
        }
  
        loop_wall=loop_wall->next;
        /* If we've gone completely around the loop and found only outers,
         then our job is done. */
        if(loop_wall->linptr==outermark){
          Done=true;
        }
      }else{
        /* Else this wall segment is alive, search for room to expand. */
        VertDex=0;
        outermark=NULL;

        /* get base point to find third */
        pdex1=ThisLine->pdex1;  pdex2=ThisLine->pdex2;
        Point1=&(points->P[pdex2]); Point2=&(points->P[pdex1]);
  
        /* find nearest to line */
        /* always searches on the left hand side of line travel direction */
        found=FindThird(pdex1,pdex2,&VertDex,&point_test);
        if(!found){
          /* Third not found, therefore this line is facing outer space. */
          ThisLine->state=outer;
          loop_wall=loop_wall->next;
        }else{
          /* Else we have a complete new triangle,
            save it and expand the loop wall. */

          /* triangle is counter-clockwise */
          store_triangle(pdex2,VertDex,pdex1);
          points->P[VertDex].ref=vertexpnt;
  
          /* Start modifying the loop wall. */
          ThisLine->state=dead;/* Remove base line from loop */
          PopLine();

          /* Add two new lines where the old one was. */
          newline=line_create(pdex1,VertDex,alive);
          PushLine(newline);

          newline=line_create(VertDex,pdex2,alive);
          PushLine(newline);
        } /* end of ThisLine on outside - else */
      }/* end if ThisLine->state==dead - else */
    }while(!Done);
  }
}/* generate_tin */

tinner_class::tinner_class()
/* *******************************************************************
*  Constructor
* ******************************************************************* */
{
	points = NULL; sort = NULL; paramv = NULL; 
	line_tree = NULL; loop_wall = NULL;
};
tinner_class::~tinner_class()
/* *******************************************************************
*  Free all of the TIN structures from memory.
* ******************************************************************* */
{
	clear();
}/* tin_destroy */

tin_bool tinner_class::tin_read_xyzs(char *filename)
/* *******************************************************************
*  Read all the xyzs from the demo xyz.txt file into our point array.
*  Returns true if successful.
* ******************************************************************* */
{
  uint32 cnt,num_points;
  char bufstr[300];
  FILE *survey;
  // XYZType *P;/* pointer for quick dereferencing. */
  tin_bool retval=true;

  double xval,yval,zval;

  // String^ fname; survey=fopen(fname,"r");
  memset(bufstr,0,300);
  
  survey=fopen(filename,"r");
  if (survey==NULL){
    retval=false;/* Error opening file. */
  }else{
    /* first line is number of data xyz points. */
    fgets(bufstr,290,survey);
    /* validate that the first line matches our xyz format. */
    if (strstr(bufstr,"num_points")!=NULL){
      sscanf(bufstr," num_points=%li\n",&num_points);

      /* allocate the xyz array to fit the data size. */
	  allocate_points(num_points);

      cnt=0;
      while(cnt<num_points && fgets(bufstr,290,survey)){
        // P=&(points->P[cnt]);
        // sscanf(bufstr," %lf %lf %lf\n",&(P->Loc[0]),&(P->Loc[1]),&(P->Loc[2]));

		sscanf(bufstr," %lf %lf %lf\n",&xval,&yval,&zval);
		set_point(cnt,xval,yval,zval);

        cnt++;
      }
      points->NumPoints=cnt;/* set the points to what was actually read, in case we came up short. */
      retval=true;
    }else{/* Doesn't match our xyz file format, lacks proper header. */
      retval=false;
    }
    fclose(survey);
  }
  return(retval);
}/* tin_read_xyzs */

void tinner_class::clear()
/* *******************************************************************
*  Free all of the structures created by loading data and TINning,
* to get ready for another round.
* ******************************************************************* */
{
	/* destructors always return NULL for convenience */
	sort=FreeSortVect(sort);
	paramv=FreeSortVect(paramv);

	/* Dispose of your own data. */
	if (points!=NULL){delete points; points=NULL;}

	loop_wall = dispose_loop(loop_wall);

	if (line_tree!=NULL){
		flexray_destroy(line_tree,&leaf_destroy);
		freesafe(line_tree); line_tree=NULL; 
	}
	if (tri_tree!=NULL){
      flexray_destroy(tri_tree,&tri_destroy);
      freesafe(tri_tree); tri_tree=NULL;
	}
    num_triangles=0;
}/* clear */

}// namespace tinlib 