using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using tinlib;

/* C sharp demo using tinlib */
namespace tinner_cs
{
    public partial class tinner_form : Form
    {
        tinner_class tinner;
        Random rand = new Random();
        /* ********************************************************************************* */
        public tinner_form()
        {/* Constructor */
            InitializeComponent();
            print_triangles_button.Enabled = false;
            maketin_button.Enabled = false;
            tinner = new tinner_class();
            this.Disposed += new EventHandler(tinner_form_Disposed);
        }/* tinner_form */
        /* ********************************************************************************* */
        void tinner_form_Disposed(object sender, EventArgs e)
        {
            tinner.Dispose();
        }
        /* ********************************************************************************* */
        private void getpoints_button_Click(object sender, EventArgs e)
        {
            print_triangles_button.Enabled = false;
            maketin_button.Enabled = true;
            create_random_points(100);/* create 100 xyz points */
            plot_box.Refresh();
        }
        /* ********************************************************************************* */
        private void create_random_points(int num_points)
        {
            /* First wipe out the old TIN points and triangles */
            tinner.clear();

            /* Always allocate the number of points you are going to need first. */
            /* This is just an array we are loading! */
            tinner.allocate_points(num_points);
            
            /* So of course the index (first parameter) must always be a 
             sequence from zero to your number of points-1 */
            for (uint cnt = 0; cnt < num_points; cnt++)
            {
                double randx = rand.Next(0, plot_box.ClientRectangle.Width);
                double randy = rand.Next(0, plot_box.ClientRectangle.Height);
                double randz = rand.NextDouble();
                tinner.set_point(cnt, randx, randy, randz);
            }
        }/* create_random_points */
        /* ********************************************************************************* */
        private void maketin_button_Click(object sender, EventArgs e)
        {
            tinner.generate_tin(1000000.0);
            plot_box.Refresh();
            print_triangles_button.Enabled = true;
        }
        /* ********************************************************************************* */
        private void plot_box_Paint(object sender, PaintEventArgs e)
        {
            Graphics grx = e.Graphics;
            plot_points(grx);
            plot_triangles(grx);
        }
        /* ********************************************************************************* */
        private void plot_points(Graphics grx)
        {/* Plot all of our xyz points. */
            int wdt = 2, hgt = 2;
            int halfwdt = wdt / 2, halfhgt = hgt / 2;
            Brush br = Brushes.Blue;
            uint num_points = tinner.get_num_points();
            for (uint point_cnt = 0; point_cnt < num_points; point_cnt++)
            {
                double x0 = 0.0, y0 = 0.0, z0 = 0.0;
                tinner.get_point(point_cnt, ref x0, ref y0, ref z0);
                grx.FillRectangle(br, (float)x0 - halfwdt, (float)y0 - halfhgt, wdt, hgt);
                // grx.FillEllipse(br, (float)x0 - halfwdt, (float)y0 - halfhgt, wdt, hgt);
            }
        }/* plot_points */
        /* ********************************************************************************* */
        private void plot_triangles(Graphics grx)
        {/* Plot all TIN triangles, if any have been created. */

            uint vdex0 = 0, vdex1 = 0, vdex2 = 0;// indexes of the points of the triangle corners
            double x0 = 0.0, y0 = 0.0, z0 = 0.0, x1 = 0.0, y1 = 0.0, z1 = 0.0, x2 = 0.0, y2 = 0.0, z2 = 0.0;

            // String outstr = String.Empty;/* initialize the output string */
            StringBuilder outstr = new StringBuilder();
            Console.WriteLine("tinner.num_triangles:{0}.", tinner.num_triangles);

            Pen pen = Pens.Red;

            /* tcnt is the triangle index */
            for (uint tdex = 0; tdex < tinner.num_triangles; tdex++)
            {
                /* v0,v1, and v2 return the indexes of each triangle corner, as found in the point array */
                if (tinner.get_triangle_indexes(tdex, ref vdex0, ref vdex1, ref vdex2) > 0)
                {
                    /* so now that we have the vertex indexes, look up the actual coordinates */
                    tinner.get_point(vdex0, ref x0, ref y0, ref z0);
                    tinner.get_point(vdex1, ref x1, ref y1, ref z1);
                    tinner.get_point(vdex2, ref x2, ref y2, ref z2);

                    /* draw the triangle itself */
                    grx.DrawLine(pen, (float)x0, (float)y0, (float)x1, (float)y1);
                    grx.DrawLine(pen, (float)x1, (float)y1, (float)x2, (float)y2);
                    grx.DrawLine(pen, (float)x2, (float)y2, (float)x0, (float)y0);
                }
                else
                {/* if it returned false, then we went out of range of the list of triangles. */
                    System.Console.WriteLine("Error, no triangle at index:{0}.", tdex);
                }
            }/* end for loop. */
            System.Console.WriteLine(outstr.ToString());
        }/* plot_triangles */
        /* ********************************************************************************* */
        private void print_triangles_button_Click(object sender, EventArgs e)
        {
            textform tf = new textform(print_triangles(tinner));
            tf.ShowDialog(this);
        }/* print_triangles_button_Click */
        /* ********************************************************************************* */
        String print_triangles(tinner_class tinner)
        /* *******************************************************************
        *  Print out all the triangles created by tinning.
        * ******************************************************************* */
        {
            uint vdex0 = 0, vdex1 = 0, vdex2 = 0;// indexes of the points of the triangle corners
            double x0 = 0.0, y0 = 0.0, z0 = 0.0, x1 = 0.0, y1 = 0.0, z1 = 0.0, x2 = 0.0, y2 = 0.0, z2 = 0.0;

            // String outstr = String.Empty;/* initialize the output string */
            StringBuilder outstr = new StringBuilder();
            Console.WriteLine("tinner.num_triangles:{0}.", tinner.num_triangles);

            /* tcnt is the triangle index */
            for (uint tdex = 0; tdex < tinner.num_triangles; tdex++)
            {
                /* v0,v1, and v2 return the indexes of each triangle corner, as found in the point array */
                if (tinner.get_triangle_indexes(tdex, ref vdex0, ref vdex1, ref vdex2) > 0)
                {
                    /* so now that we have the vertex indexes, look up the actual coordinates */
                    tinner.get_point(vdex0, ref x0, ref y0, ref z0);
                    tinner.get_point(vdex1, ref x1, ref y1, ref z1);
                    tinner.get_point(vdex2, ref x2, ref y2, ref z2);

                    /* now just put the results into the output string */
                    outstr.Append(System.String.Format("triangle {0}: ", tdex));
                    outstr.Append(System.String.Format("(x:{0} y:{1} z:{2}) ", x0, y0, z0));
                    outstr.Append(System.String.Format("(x:{0} y:{1} z:{2}) ", x1, y1, z1));
                    outstr.Append(System.String.Format("(x:{0} y:{1} z:{2}) ", x2, y2, z2));
                    outstr.Append(Environment.NewLine);
                }
                else
                {/* if it returned false, then we went out of range of the list of triangles. */
                    System.Console.WriteLine("Error, no triangle at index:{0}.", tdex);
                }
            }/* end for loop. */
            return outstr.ToString();
            // this.output_textbox.Text = outstr; System.Console.WriteLine(outstr.ToString());
        }/* print_triangles */
        /* ********************************************************************************* */
        public class textform : Form
        {/* For quick display of triangles as text coordinates */
            public TextBox tex = new TextBox();
            public textform(String contents)
            {
                this.Width = 1000; this.Height = 1000;
                tex = new TextBox();
                tex.Multiline = true;
                tex.Dock = DockStyle.Fill;
                this.Controls.Add(tex);
                tex.Text = contents;
                tex.Show();
            }
        }
        /* ********************************************************************************* */
    }
}
